>module Pprint (
>  commaConcat, dotConcat, termConcat, nlConcat, commaNlConcat,
>  opConcat, format, heading, makeComment,
>  makeLogicalLines, makeUnion, makeUnionOneLine, maybeString
>  )

>where

>import Useful

============================================================================
Printing functions

>maybeString :: Bool -> String -> String
>maybeString True st = st
>maybeString False _ = ""

>colwidth :: Int
>colwidth = 75

Print main heading

>heading h = 
>  let w = length h
>      s1 = (colwidth - w - 5) `div` 2
>      s2 = colwidth - w - 5 - s1
>  in "-- " ++ copy '*' (colwidth -3) ++ "\n" ++ 
>     "-- *" ++ spaces s1 ++ h ++ spaces s2 ++ "*\n" ++
>     "-- " ++ copy '*' (colwidth -3) ++ "\n\n"

Turn string into comment, by splitting into lines of length at most 
colwidth-3, and prefixing each line with "-- "

>makeComment :: String -> String
>makeComment st = 
>  let lines = splitLines "" "" (words st)
>  in  lines

Split list of words into lines, accumulating current line in l, and complete
result in st

>splitLines :: String -> String -> [String] -> String 
>splitLines l st [] = st++makeC l
>splitLines l st (w:ws) =
>  if length l+length w+4 > colwidth 
>  then splitLines  "" (st++makeC l) (w:ws)
>  else splitLines (l++" "++w) st ws

>makeC l = "-- "++l++"\n"

>commaConcat [] = ""
>commaConcat sts = foldr1 comma sts
> where comma xs ys = xs ++ ", " ++ ys

>dotConcat = foldr1 dot
> where dot xs ys = xs ++ "." ++ ys

>nlConcat n = foldr1 nl
> where nl xs ys = xs ++ "\n" ++ copy ' ' n ++ ys

>commaNlConcat n = foldr1 commaNl
> where commaNl xs ys = xs ++ ",\n" ++ copy ' ' n ++ ys

>termConcat = foldr1 term
> where term xs ys = xs ++ " | " ++ ys

>termnlConcat = foldr1 termnl
> where termnl xs ys = xs ++ " |\n" ++ ys

>opConcat _ [] = ""
>opConcat op sts = foldr1 (f op) sts
> where f op xs ys = xs ++ op ++ ys

Format list of strings to fit in column of width n, separating items with
separator


>prettyPrint :: Int -> String -> [String] -> [String]
>prettyPrint n separator [] = []
>prettyPrint n separator (st:sts) = pp st sts
> where pp line [] = [line]
>       pp line (st:sts)
>         | length line + length st > n  
>                        = (line ++ separator ++ "\n") : pp st sts
>         | otherwise    = pp (line ++ separator ++ st) sts

Print each string, prefixed by n spaces

>prefixStrings :: Int -> [String] -> String
>prefixStrings n [] = ""
>prefixStrings n sts = foldr1 foo sts
>  where foo st st' = st ++ copy ' ' n ++ st'

>format :: Int -> String -> [String] -> String
>format indent separator = 
>  prefixStrings indent . 
>  prettyPrint (colwidth - indent - length separator) separator

================= Misc things, that probably shouldn't be here

print union, indenting by n spaces

>makeUnion :: Int -> [String] -> String
>makeUnion _ [] = "{}\n"
>makeUnion n [a] = 
>  (if length a > colwidth-25 then "\n"++copy ' ' n else "")++a++"\n"
>makeUnion n as
>  | longline  = "\n"++spaces n++"Union({\n"++spaces (n+2)++
>                commaNlConcat (n+2) as++"\n"++ spaces n++"})\n"
>  | otherwise = "Union({"++commaConcat as++"})\n"
>  where
>        longline = 2 * length as + sum (map length as) + n + 20 > colwidth

>makeUnionOneLine :: [String] -> String
>makeUnionOneLine [] = "{}"
>makeUnionOneLine [a] = a
>makeUnionOneLine [a,b] = "union("++a++", "++b++")"
>makeUnionOneLine as = "Union({"++commaConcat as++"})"

Combine lines where first ends in "\"

>makeLogicalLines [] = []
>makeLogicalLines [l] = [l]
>makeLogicalLines (l1:l2:ls)
>  | l1 /= "" && last l1 == '\\'   = (init l1 ++ hd ls') : tl ls'
>  | otherwise                     = l1 : ls'
>  where ls' = makeLogicalLines (l2:ls)
