/*
 * Created on Jun 16, 2003
 */
package core;

import java.io.Serializable;
import java.security.Key;
import java.security.NoSuchAlgorithmException;

/**
 * @author Xavier Didelot
 */
public interface IMessage extends Serializable {
	/**
	 * This method returns the value of the protected member string
	 * It is especially useful to send a message to another host
	 * @return the content of the message
	 */
	public abstract String getString();
	/**
	 * Change the value of the private member string into the given string
	 * It is advised to only use Base64 string in the message to avoid any issue 
	 * @param string new value for the content of the message
	 */
	public abstract void setString(String string);
	/**
	 * This method concats two messages and put an 'at' symbol between them so that we can
	 * divide it later with the method getAllFields for example
	 * This means that a Message should never contain the symbol 'at'
	 * It is advised to only use Base64 string in the message to avoid any issue 
	 * @param msg
	 */
	public abstract void addMessage(Message msg);
	/**
	 * Returns an array containing the different parts of a message, ie the different
	 * messages that have been concatenated in order to build this message
	 * @return an array containing the different parts of a message
	 */
	public abstract IMessage[] getAllFields();
	/**
	 * Encrypts the message with the given key, using the given algorithm, and put the result in Base64 format
	 * @param algorithm algorithm to use for encryption
	 * @param key key to use for encryption
	 * @throws Exception
	 */
	public abstract void encrypt(String algorithm, Key key) throws Exception;
	/**
	 * Decrypt the message with a given key, using the given algorithm. The message has to be in Base64 format.
	 * @param algorithm algorithm to use for decryption
	 * @param key key to use for decryption
	 * @throws Exception
	 */
	public abstract void decrypt(String algorithm, Key key) throws Exception;
	/**
	 * This method hashes the Message
	 * @throws NoSuchAlgorithmException
	 */
	public abstract void hash(String method) throws NoSuchAlgorithmException;

	public abstract void encrypt(String algorithm, Key key,String protoTag) throws Exception;
	public abstract void decrypt(String algorithm, Key key, String protoTag) throws Exception;
	public abstract void hash(String method,String protoTag) throws NoSuchAlgorithmException;
	/**
	 * Check if the Message is equal to another one
	 * @param other the other Message
	 * @return true if the messages are the same, false otherwise
	 */
	public abstract boolean equals(Message other);
	/**
	 * This method XORes the last two fields of the  Message
	 * @throws Exception
	 */
	public abstract void xor() throws Exception;
	/**
	 * This method XORes the current Message with a given Message
	 * @param message
	 */
	public abstract void xor(IMessage message);	
	public abstract void setStringAddr(String string); 
	public abstract void unxor(IMessage message);
}