package protocols;

import core.*;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.lang.reflect.Array;
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.math.BigInteger;

public class RESPOR extends Protocol {
   
   public RESPOR (ObjectOutputStream oos) {
      super("core.MessageT","core.MessageKeyT","core.NonceT");
      this.oos=oos;
      }
   
   public void run() {
      try {
         IMessageKey kab = null;
         
         println("creating nonce nb");
         createNewNonce("nb");
         
         println("waiting for m1a");
         Message m1a = message(get("a",4002));
         checkTag(m1a,"AAN");
         addParts(m1a.getAllFields(),3);
         m0=getPart();
         setNameToId(m0.getString(),"a");
         m0=getPart();
         setNameToId(m0.getString(),"b");
         m0=getPart();
         setValueToNonce(m0.getString(),"m");
         
         
         println("Building message m2");
         Message m2 =message();
         m2.addMessage(message(nonceFromId("m")));
         m2.addMessage(message(agentFromId("a")));
         m2.addMessage(message(agentFromId("b")));
         m0=message();
         m0.addMessage(message(nonceFromId("nb")));
         m0.addMessage(message(nonceFromId("m")));
         m0.addMessage(message(agentFromId("a")));
         m0.addMessage(message(agentFromId("b")));
         println("Encrypting m0");
         m0.encrypt("DES", getSecretKey(nameFromId("b")));
         m2.addMessage(m0);
         println("Sending message m2");
         put("s", m2.getString(),4001);
         
         println("waiting for m3");
         Message m3 = message(get("s"));
         checkTag(m3,"NKNK)");
         addParts(m3.getAllFields(),2);
         m0=getPart();
         setValueToNonce(m0.getString(),"m");
         m0=getPart();
         println("Decrypting m0");
         m0.decrypt("DES", getSecretKey(nameFromId("b")));
         checkTag(m0,"NK");
         addParts(m0.getAllFields(),2);
         m0=getPart();
         setValueToNonce(m0.getString(),"nb");
         m0=getPart();
         kab= messageKey(m0);
         
         
         
         
         } catch (Exception e) {e.printStackTrace();System.exit(0);}
      }
   
   public static void subMain(String[] args,ObjectOutputStream oos) {
      RESPOR b;
      while (true) {
         b=new RESPOR(oos);
         args=b.getArgs(args);
         b.newAgent("a",4002);
         b.start();
         }
      }
   
   public String[] getArgs(String[] args) {
      try {
         if (args.length != 4) {
            args = (String[]) Array.newInstance("".getClass(), 4);
            BufferedReader in2 =
            new BufferedReader(new InputStreamReader(System.in));
            println("Type the SecretKeyStore file to use [secretkeystore]");
            args[0]=in2.readLine();
            if (args[0].compareTo("")==0) args[0]="secretkeystore";
            println("Type the DNS name of agent s [localhost]");
            args[1]=in2.readLine();
            if (args[1].compareTo("")==0) args[1]="localhost";
            println("Type the name of agent b [b]");
            args[2]=in2.readLine();
            if (args[2].compareTo("")==0) args[2]="b";
            println("Type the name of agent s [s]");
            args[3]=in2.readLine();
            if (args[3].compareTo("")==0) args[3]="s";
            }
         setSKS(args[0]);
         setIPToId(args[1],"s");
         setNameToId(args[2],"b");
         setNameToId(args[3],"s");
         } catch (Exception e) {e.printStackTrace();}
      return args;
      }
   
   public static void main(String[] args) {
      ObjectOutputStream oos=null;
      try {
         ByteArrayOutputStream os = new ByteArrayOutputStream();
         oos=new ObjectOutputStream(os);
         PrintOOS p=new PrintOOS("RESPOR",oos,os);
         p.start();}
      catch (Exception e) {e.printStackTrace();}
      subMain(args,oos);
      }
   }